from .document import document_schema
from .gallery import gallery_schema
from .presentation import presentation_schema
from .super_editor_document import super_editor_document_schema

__all__ = [
    document_schema,
    gallery_schema,
    presentation_schema,
    super_editor_document_schema,
]



from abc import ABC, abstractmethod
from .agent import AgentChatContext
from jsonschema import validate
from meshagent.tools import Response, Toolkit, ToolkitBuilder, ToolkitConfig
from meshagent.api import RoomClient, RoomException, RemoteParticipant
from typing import Any, Optional, Callable, TypeVar, Generic, Literal

TEvent = TypeVar("T")


class ToolResponseAdapter(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def to_plain_text(self, *, room: RoomClient, response: Response):
        pass

    @abstractmethod
    async def create_messages(
        self,
        *,
        context: AgentChatContext,
        tool_call: Any,
        room: RoomClient,
        response: Response,
    ) -> list:
        pass


class LLMAdapter(Generic[TEvent]):
    @abstractmethod
    def default_model(self) -> str: ...

    def create_chat_context(self) -> AgentChatContext:
        return AgentChatContext()

    async def check_for_termination(
        self, *, context: AgentChatContext, room: RoomClient
    ):
        return True

    def tool_providers(self, *, model: str) -> list[ToolkitBuilder]:
        return []

    async def make_toolkit(
        self, *, room: RoomClient, model: str, config: ToolkitConfig
    ) -> Toolkit:
        for tool in self.tool_providers(model=model):
            if tool.name == config.name:
                return Toolkit(
                    name=config.name,
                    tools=[await tool.make(room=room, model=model, config=config)],
                )

        raise RoomException(f"Unexpected tool: {config.name} for model {model}")

    @abstractmethod
    async def next(
        self,
        *,
        context: AgentChatContext,
        room: RoomClient,
        toolkits: list[Toolkit],
        tool_adapter: Optional[ToolResponseAdapter] = None,
        output_schema: Optional[dict] = None,
        event_handler: Optional[Callable[[TEvent], None]] = None,
        model: Optional[str] = None,
        on_behalf_of: Optional[RemoteParticipant] = None,
    ) -> Any:
        pass

    def validate(response: dict, output_schema: dict):
        validate(response, output_schema)


class MessageStreamLLMAdapter(LLMAdapter):
    def __init__(
        self, *, participant_name: str, context_mode: Literal["diff", "full"] = "diff"
    ):
        self.participant_name = participant_name
        self.context_mode = context_mode

    def default_model(self) -> str:
        return "toolkit"

    def create_chat_context(self) -> AgentChatContext:
        return AgentChatContext()

    async def check_for_termination(
        self, *, context: AgentChatContext, room: RoomClient
    ):
        return True

    async def next(
        self,
        *,
        context: AgentChatContext,
        room: RoomClient,
        toolkits: list[Toolkit],
        tool_adapter: Optional[ToolResponseAdapter] = None,
        output_schema: Optional[dict] = None,
        event_handler: Optional[Callable[[TEvent], None]] = None,
        model: Optional[str] = None,
        on_behalf_of: Optional[RemoteParticipant] = None,
    ) -> Any:
        participant = room.messaging.get_participant_by_name(self.participant_name)
        if participant is None:
            raise RoomException("participant is not currently connected")

        stream = await room.messaging.create_stream(
            to=participant,
            header={
                "context": context.to_json(),
                "model": model,
                "output_schema": output_schema,
                "on_behalf_of_id": on_behalf_of.id if on_behalf_of else None,
                "metadata": context.metadata,
            },
        )

        error = None
        output = None
        try:
            async for chunk in stream.read_chunks():
                event = chunk.header.get("event")
                if event is not None and event_handler is not None:
                    event_handler(event)

                output = chunk.header.get("output")
                if output is not None:
                    output.append(output)

                if chunk.header.get("done"):
                    break

        except Exception as ex:
            error = ex

        await stream.close()

        if self.context_mode == "diff":
            context.messages.clear()

        if error:
            raise error

        return output

    def validate(response: dict, output_schema: dict):
        validate(response, output_schema)



from meshagent.api.schema import MeshSchema, ElementType, ChildProperty, ValueProperty

widget_names = {
    "button",
    "box",
    "canvas",
    "text",
    "image",
    "video",
    "hstack",
    "vstack",
    "zstack",
    "input",
    "textarea",
    "form",
    "select",
    "checkbox",
    "radiogroup",
    "spacer",
    "divider",
    "spinner",
    "markdown",
    "datepicker",
    "scrollview",
    "format",
    "datagrid",
    "chat",
    "tabs",
}


def text_style_properties():
    return [
        ValueProperty(
            name="color", description="the color of the text #rrggbbaa", type="string"
        ),
        ValueProperty(
            name="emphasis",
            description="the emphasis for the text",
            type="string",
            enum=["none", "italic"],
        ),
        ValueProperty(
            name="decoration",
            description="the decoration for the text",
            type="string",
            enum=["none", "underline", "strikethrough"],
        ),
        ValueProperty(
            name="font", description="the name of a font to render with", type="string"
        ),
        ValueProperty(
            name="weight",
            description="the font weight to use",
            type="string",
            enum=[
                "bold",
                "normal",
                "100",
                "200",
                "300",
                "400",
                "500",
                "600",
                "700",
                "800",
                "900",
            ],
        ),
        ValueProperty(
            name="align",
            description="the text alignment to use",
            type="string",
            enum=["left", "right", "center", "none"],
        ),
        ValueProperty(
            name="letterSpacing", description="the letter spacing to use", type="number"
        ),
        ValueProperty(name="lineHeight", description="the line height", type="number"),
        ValueProperty(name="size", description="the text size", type="number"),
    ]


def box_properties():
    return [
        ValueProperty(
            name="topLeftRadius",
            description="the top left radius for the box",
            type="number",
        ),
        ValueProperty(
            name="topRightRadius",
            description="the top right radius for the box",
            type="number",
        ),
        ValueProperty(
            name="bottomLeftRadius",
            description="the bottom left radius for the box",
            type="number",
        ),
        ValueProperty(
            name="bottomRightRadius",
            description="the bottom right radius for the box",
            type="number",
        ),
        ValueProperty(
            name="marginTop", description="the top margin for the box", type="number"
        ),
        ValueProperty(
            name="marginBottom",
            description="the bottom margin for the box",
            type="number",
        ),
        ValueProperty(
            name="marginLeft", description="the left margin for the box", type="number"
        ),
        ValueProperty(
            name="marginRight",
            description="the right margin for the box",
            type="number",
        ),
        ValueProperty(
            name="paddingTop", description="the top padding for the box", type="number"
        ),
        ValueProperty(
            name="paddingBottom",
            description="the bottom padding for the box",
            type="number",
        ),
        ValueProperty(
            name="paddingLeft",
            description="the left padding for the box",
            type="number",
        ),
        ValueProperty(
            name="paddingRight",
            description="the right padding for the box",
            type="number",
        ),
        ValueProperty(
            name="borderColor",
            description="the border color for the box",
            type="number",
        ),
        ValueProperty(
            name="borderTop", description="the top border for the box", type="number"
        ),
        ValueProperty(
            name="borderBottom",
            description="the bottom border for the box",
            type="number",
        ),
        ValueProperty(
            name="borderLeft", description="the left border for the box", type="number"
        ),
        ValueProperty(
            name="borderRight",
            description="the right border for the box",
            type="number",
        ),
        ValueProperty(
            name="borderTopColor",
            description="the top border color for the box #rrggbbaa",
            type="string",
        ),
        ValueProperty(
            name="borderBottomColor",
            description="the bottom border color for the box #rrggbbaa",
            type="string",
        ),
        ValueProperty(
            name="borderLeftColor",
            description="the left border color for the box #rrggbbaa",
            type="string",
        ),
        ValueProperty(
            name="borderRightColor",
            description="the right border color for the box #rrggbbaa",
            type="string",
        ),
        ValueProperty(
            name="backgroundColor",
            description="the background color for the box #rrggbbaa",
            type="string",
        ),
        ValueProperty(
            name="backgroundImageUrl", description="a background image", type="string"
        ),
        ValueProperty(
            name="backgroundImageFit",
            description="how to position the background image",
            type="string",
            enum=[
                "fit",
                "contain",
                "cover",
                "fitWidth",
                "fitHeight",
                "scaleDown",
                "none",
            ],
        ),
        ValueProperty(
            name="linearGradient",
            description="a linear gradient for the background",
            type="string",
        ),
        ValueProperty(
            name="radialGradient",
            description="a radial gradient for the background",
            type="string",
        ),
        ValueProperty(name="width", description="the width for the box", type="number"),
        ValueProperty(
            name="height", description="the height for the box", type="number"
        ),
    ]


def widget_properties():
    return [
        ValueProperty(
            name="name",
            description="the name of the widget, must be unique in this widget",
            type="string",
        ),
        ValueProperty(
            name="script",
            description="a script that contains the logic for this button",
            type="string",
        ),
        ValueProperty(
            name="visible",
            description="the whether the widget is initially visible",
            type="boolean",
        ),
    ]


def child_widgets():
    return [
        ChildProperty(
            name="children",
            description="child widgets",
            child_tag_names=widget_names,
        )
    ]


button = ElementType(
    tag_name="button",
    description="a button",
    properties=[
        *widget_properties(),
        *box_properties(),
        *child_widgets(),
    ],
)

box = ElementType(
    tag_name="box",
    description="a box",
    properties=[
        *widget_properties(),
        *box_properties(),
        *child_widgets(),
    ],
)


canvas = ElementType(
    tag_name="canvas",
    description="a canvas",
    properties=[
        *widget_properties(),
        *box_properties(),
        *child_widgets(),
    ],
)

text = ElementType(
    tag_name="text",
    description="text",
    properties=[
        ValueProperty(name="text", description="the text to display", type="string"),
        *widget_properties(),
        *text_style_properties(),
    ],
)

markdown = ElementType(
    tag_name="markdown",
    description="markdown",
    properties=[
        *widget_properties(),
        ValueProperty(
            name="text", description="the markdown to display", type="string"
        ),
    ],
)

image = ElementType(
    tag_name="image",
    description="image",
    properties=[
        *widget_properties(),
        ValueProperty(name="url", description="the url of an image", type="string"),
        ValueProperty(
            name="alt", description="a description of the image", type="string"
        ),
        ValueProperty(
            name="fit",
            description="how to position the image",
            type="string",
            enum=[
                "fit",
                "contain",
                "cover",
                "fitWidth",
                "fitHeight",
                "scaleDown",
                "none",
            ],
        ),
        *box_properties(),
    ],
)

video = ElementType(
    tag_name="video",
    description="video",
    properties=[
        *widget_properties(),
        ValueProperty(name="url", description="the url of a video", type="string"),
        ValueProperty(
            name="alt", description="a description of the video", type="string"
        ),
        ValueProperty(
            name="fit",
            description="how to position the video",
            type="string",
            enum=[
                "fit",
                "contain",
                "cover",
                "fitWidth",
                "fitHeight",
                "scaleDown",
                "none",
            ],
        ),
        *box_properties(),
    ],
)

scrollview = ElementType(
    tag_name="scrollview",
    description="scrollview",
    properties=[
        *widget_properties(),
        *box_properties(),
        ValueProperty(
            name="spacing", description="spacing between the items", type="number"
        ),
        ValueProperty(
            name="direction",
            description="spacing between the items",
            type="string",
            enum=["horizontal", "vertical"],
            required=True,
        ),
        *child_widgets(),
    ],
)


hstack = ElementType(
    tag_name="hstack",
    description="hstack",
    properties=[
        *widget_properties(),
        ValueProperty(
            name="spacing", description="spacing between the items", type="number"
        ),
        *child_widgets(),
    ],
)

vstack = ElementType(
    tag_name="vstack",
    description="vstack",
    properties=[
        ValueProperty(
            name="spacing", description="spacing between the items", type="number"
        ),
        *widget_properties(),
        *child_widgets(),
    ],
)

zstack = ElementType(
    tag_name="zstack",
    description="zstack",
    properties=[
        *widget_properties(),
        *child_widgets(),
    ],
)

form = ElementType(
    tag_name="form",
    description="form",
    properties=[
        *widget_properties(),
        *child_widgets(),
    ],
)

input = ElementType(
    tag_name="input",
    description="input",
    properties=[
        ValueProperty(
            name="placeholder",
            description="a placeholder for the field",
            type="string",
        ),
        ValueProperty(
            name="initialValue",
            description="the initial value of the input field",
            type="string",
        ),
        ValueProperty(
            name="obscure",
            description="whether to obscure the input",
            type="boolean",
        ),
        *widget_properties(),
        *box_properties(),
        *text_style_properties(),
    ],
)

select = ElementType(
    tag_name="select",
    description="select",
    properties=[
        ValueProperty(
            name="placeholder",
            description="a placeholder for the field",
            type="string",
        ),
        ValueProperty(
            name="initialValue",
            description="the initial value of the input field",
            type="string",
        ),
        ChildProperty(name="options", child_tag_names=["option"]),
        *widget_properties(),
        *box_properties(),
    ],
)

textarea = ElementType(
    tag_name="textarea",
    description="textarea",
    properties=[
        ValueProperty(
            name="placeholder",
            description="a placeholder for the field",
            type="string",
        ),
        ValueProperty(
            name="initialValue",
            description="the initial value of the input field",
            type="string",
        ),
        *widget_properties(),
        *text_style_properties(),
        *box_properties(),
    ],
)

checkbox = ElementType(
    tag_name="checkbox",
    description="checkbox",
    properties=[
        ValueProperty(
            name="initialValue",
            description="the initial value of the input field",
            type="string",
            enum=["checked", "unchecked"],
        ),
        *child_widgets(),
    ],
)

datepicker = ElementType(
    tag_name="datepicker",
    description="datepicker",
    properties=[
        ValueProperty(
            name="initialValue",
            description="the initial value of the input field",
            type="string",
        ),
        *widget_properties(),
        *box_properties(),
        *text_style_properties(),
    ],
)

option = ElementType(
    tag_name="option",
    description="option",
    properties=[
        ValueProperty(
            name="value",
            description="the initial value of the input field",
            type="string",
        ),
        *child_widgets(),
    ],
)

radio = ElementType(
    tag_name="radio",
    description="radio",
    properties=[
        ValueProperty(
            name="value",
            description="the initial value of the input field",
            type="string",
        ),
        *child_widgets(),
    ],
)

radiogroup = ElementType(
    tag_name="radiogroup",
    description="radiogroup",
    properties=[
        ValueProperty(
            name="initialValue",
            description="the initial value of the input field",
            type="string",
        ),
        ChildProperty(name="radios", child_tag_names=["radio"]),
        *widget_properties(),
    ],
)

spacer = ElementType(
    tag_name="spacer",
    description="spacer",
    properties=[
        *widget_properties(),
    ],
)

divider = ElementType(
    tag_name="divider",
    description="divider",
    properties=[
        ValueProperty(
            name="color",
            description="the color of the divider #rrggbbaa",
            type="string",
        ),
        ValueProperty(name="marginTop", description="the top margin", type="number"),
        ValueProperty(
            name="marginBottom", description="the bottom margin", type="number"
        ),
        ValueProperty(name="marginLeft", description="the left margin", type="number"),
        ValueProperty(
            name="marginRight", description="the right margin", type="number"
        ),
        *widget_properties(),
    ],
)

spinner = ElementType(
    tag_name="spinner",
    description="spinner",
    properties=[
        ValueProperty(
            name="color",
            description="the color of the spinner #rrggbbaa",
            type="string",
        ),
        ValueProperty(
            name="width", description="the height of the spinner", type="number"
        ),
        ValueProperty(
            name="height", description="the width of the spinner", type="number"
        ),
        *widget_properties(),
    ],
)

widgets = ElementType(
    tag_name="widgets",
    description="widgets",
    properties=[*child_widgets()],
)

widget = ElementType(
    tag_name="widget",
    description="widget",
    properties=[
        ChildProperty(
            name="children",
            description="child widgets",
            ordered=True,
            child_tag_names=[
                "widgets",
            ],
        )
    ],
)


chat = ElementType(
    tag_name="chat",
    description="widget",
    properties=[
        ValueProperty(
            name="path",
            description="the path to the chat thread",
            type="string",
        ),
        ValueProperty(
            name="agent",
            description="the name of an agent to chat with on the thread",
            type="string",
        ),
        ValueProperty(
            name="initalStyle",
            description="how should the chat look before the user interacts with the agent",
            type="string",
            enum=["centered", "feed"],
        ),
        *box_properties(),
    ],
)


gridcolumn = ElementType(
    tag_name="gridcolumn",
    description="data grid column",
    properties=[
        ValueProperty(
            name="label",
            description="a friendly label for the column",
            type="string",
        ),
        ValueProperty(
            name="name",
            description="the name of the column",
            type="string",
        ),
        ValueProperty(
            name="width",
            description="the width of the column",
            type="number",
        ),
        ValueProperty(
            name="height",
            description="the width of the column",
            type="number",
        ),
        ValueProperty(
            name="cursor",
            description="the width of the column",
            type="string",
            enum=["none", "click"],
        ),
        *text_style_properties(),
    ],
)


datagrid = ElementType(
    tag_name="datagrid",
    description="data grid",
    properties=[
        *widget_properties(),
        *box_properties(),
        ValueProperty(
            name="table",
            description="the table to query for data",
            type="string",
        ),
        ValueProperty(
            name="filter",
            description="a filter statement",
            type="string",
        ),
        ChildProperty(
            name="columns",
            description="columns in the grid",
            ordered=True,
            child_tag_names=[
                "gridcolumn",
            ],
        ),
    ],
)

tab = ElementType(
    tag_name="tab",
    description="tab",
    properties=[
        *box_properties(),
        *widget_properties(),
        *child_widgets(),
        ValueProperty(
            name="activeBackgroundColor",
            description="the color of the background when the tab is active #rrggbbaa",
            type="string",
        ),
    ],
)

tabs = ElementType(
    tag_name="tabs",
    description="tabs",
    properties=[
        *box_properties(),
        *widget_properties(),
        ValueProperty(
            name="spacing", description="spacing between the items", type="number"
        ),
        ValueProperty(
            name="initialTab", description="the initially selected tab", type="string"
        ),
        ChildProperty(
            name="tabs",
            description="columns in the grid",
            ordered=True,
            child_tag_names=[
                "tab",
            ],
        ),
    ],
)


format = ElementType(
    tag_name="format",
    description="format text using an expression",
    properties=[
        *text_style_properties(),
        *widget_properties(),
    ],
)

widget_schema = MeshSchema(
    root_tag_name="widget",
    elements=[
        widget,
        canvas,
        box,
        button,
        text,
        video,
        image,
        hstack,
        vstack,
        zstack,
        input,
        textarea,
        form,
        select,
        checkbox,
        radiogroup,
        datepicker,
        spacer,
        divider,
        spinner,
        markdown,
        radio,
        option,
        widgets,
        scrollview,
        datagrid,
        gridcolumn,
        format,
        chat,
        tabs,
        tab,
    ],
)

if __name__ == "__main__":
    import json

    print(json.dumps(widget_schema.to_json()))



from meshagent.api.schema import ValueProperty, ChildProperty
from meshagent.tools import Toolkit
from meshagent.agents.writer import Writer, WriterContext
from meshagent.agents.adapter import LLMAdapter, ToolResponseAdapter
from meshagent.api.schema_util import prompt_schema, merge
from typing import Optional
from meshagent.api import Requirement

import logging

logger = logging.getLogger("planning_agent")


class SingleShotWriter(Writer):
    def __init__(
        self,
        name: str,
        llm_adapter: LLMAdapter,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        description: Optional[str] = None,
        title: Optional[str] = None,
        rules: Optional[list[str]] = None,
        requires: Optional[list[Requirement]] = None,
        toolkits: Optional[list[Toolkit]] = None,
        supports_tools: Optional[bool] = None,
        labels: Optional[list[str]] = None,
    ):
        super().__init__(
            name=name,
            description=description,
            title=title,
            input_schema=merge(
                schema=prompt_schema(description="use a prompt to generate content"),
                additional_properties={"path": {"type": "string"}},
            ),
            output_schema={
                "type": "object",
                "additionalProperties": False,
                "required": [],
                "properties": {},
            },
            requires=requires,
            supports_tools=supports_tools,
        )
        self._rules = rules
        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter
        if toolkits is None:
            toolkits = []
        self._toolkits = toolkits
        self._labels = labels

    async def init_chat_context(self):
        context = self._llm_adapter.create_chat_context()
        context.append_rules(self._rules)
        return context

    async def write(self, writer_context: WriterContext, arguments: dict):
        arguments = arguments.copy()
        self.pop_path(arguments=arguments)

        prompt = arguments["prompt"]

        writer_context.call_context.chat.append_user_message(message=prompt)

        toolkits = [*self._toolkits, *writer_context.call_context.toolkits]

        try:
            response = await self._llm_adapter.next(
                context=writer_context.call_context.chat,
                room=writer_context.room,
                toolkits=toolkits,
                tool_adapter=self._tool_adapter,
                output_schema=writer_context.document.schema.to_json(),
            )

        except Exception as e:
            logger.error("Unable to execute reasoning completion task", exc_info=e)
            # retry
            raise (e)

        document = writer_context.document
        response = response[document.schema.root.tag_name]

        for p in document.schema.root.properties:
            if isinstance(p, ValueProperty):
                document.root[p.name] = response[p.name]
            elif isinstance(p, ChildProperty):
                for value in response[p.name]:
                    document.root.append_json(value)
            else:
                raise Exception("Unexpected property type")

        return {}



from typing import Optional
import json
from meshagent.api.messaging import unpack_message, pack_message
from meshagent.api.room_server_client import (
    RoomException,
    RequiredToolkit,
    Requirement,
    RequiredSchema,
    RequiredTable,
)
from meshagent.api import (
    ToolDescription,
    ToolkitDescription,
    Participant,
    RemoteParticipant,
    StorageEntry,
)
from meshagent.api.protocol import Protocol
from meshagent.tools import (
    Toolkit,
    Tool,
    ToolContext,
)
from meshagent.api.room_server_client import RoomClient
from jsonschema import validate
from .context import AgentCallContext, AgentChatContext
from meshagent.api.schema_util import no_arguments_schema
import logging
import asyncio

logger = logging.getLogger("agent")


class AgentException(RoomException):
    pass


class RoomTool(Tool):
    def __init__(
        self,
        *,
        toolkit_name: str,
        name,
        input_schema,
        title=None,
        description=None,
        rules=None,
        thumbnail_url=None,
        participant_id: Optional[str] = None,
        on_behalf_of_id: Optional[str] = None,
        defs: Optional[dict] = None,
    ):
        self._toolkit_name = toolkit_name
        self._participant_id = participant_id
        self._on_behalf_of_id = on_behalf_of_id

        super().__init__(
            name=name,
            input_schema=input_schema,
            title=title,
            description=description,
            rules=rules,
            thumbnail_url=thumbnail_url,
            defs=defs,
        )

    async def execute(self, context, **kwargs):
        return await context.room.agents.invoke_tool(
            toolkit=self._toolkit_name,
            tool=self.name,
            participant_id=self._participant_id,
            on_behalf_of_id=self._on_behalf_of_id,
            arguments=kwargs,
        )


class Agent:
    def __init__(
        self,
        *,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        requires: Optional[list[Requirement]] = None,
        labels: Optional[list[str]] = None,
        skills_dirs: Optional[list[str]] = None,
    ):
        self._name = name
        if title is None:
            title = name
        self._title = title
        if description is None:
            description = ""

        self._description = description
        if requires is None:
            requires = []

        self._requires = requires

        if labels is None:
            labels = []

        self._labels = labels
        self._skills_dirs = None

    def get_requirements(self) -> list[Requirement]:
        return self._requires

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def title(self):
        return self._title

    @property
    def requires(self):
        return self._requires

    @property
    def labels(self):
        return self._labels

    async def init_chat_context(self) -> AgentChatContext:
        return AgentChatContext()

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "requires": list(map(lambda x: x.to_json(), self.requires)),
            "labels": self.labels,
        }


async def install_required_table(*, room: RoomClient, table: RequiredTable):
    await room.database.create_table_with_schema(
        name=table.name,
        mode="create_if_not_exists",
        schema=table.schema,
        namespace=table.namespace,
    )

    indexes = await room.database.list_indexes(
        table=table.name, namespace=table.namespace
    )

    def index_exists(column: str):
        for i in indexes:
            if column in i.columns:
                return True

        return False

    for vi in table.vector_indexes or []:
        if not index_exists(vi):
            try:
                await room.database.create_vector_index(
                    table=table.name,
                    column=vi,
                    namespace=table.namespace,
                    replace=True,
                )
            except Exception as e:
                logger.warning(f"unable to create vector index {e}", exec_info=e)

    for ti in table.full_text_search_indexes or []:
        if not index_exists(ti):
            try:
                await room.database.create_full_text_search_index(
                    table=table.name,
                    column=ti,
                    namespace=table.namespace,
                    replace=True,
                )
            except Exception as e:
                logger.warning(
                    f"unable to create full text search index {e}",
                    exec_info=e,
                )

    for si in table.scalar_indexes or []:
        if not index_exists(si):
            try:
                await room.database.create_scalar_index(
                    table=table.name,
                    column=si,
                    namespace=table.namespace,
                    replace=True,
                )
            except Exception as e:
                logger.warning(f"unable to create scalar index {e}", exec_info=e)

    logger.info(f"optimizing table {table.name} in {table.namespace}")

    # TODO: use index_stats to determine when indexes need to be updated
    await room.database.optimize(table=table.name, namespace=table.namespace)


class SingleRoomAgent(Agent):
    def __init__(
        self,
        *,
        name,
        title=None,
        description=None,
        requires=None,
        labels: Optional[list[str]] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            labels=labels,
        )
        self._room = None

    async def start(self, *, room: RoomClient) -> None:
        if self._room is not None:
            raise RoomException("agent is already started")

        self._room = room

        await self.install_requirements()

    async def stop(self) -> None:
        self._room = None
        pass

    @property
    def room(self):
        return self._room

    async def install_requirements(self, participant_id: Optional[str] = None):
        schemas_by_name = dict[str, StorageEntry]()
        toolkits_by_name = dict[str, ToolkitDescription]()

        async def refresh_schemas():
            schemas = await self._room.storage.list(path=".schemas")

            for schema in schemas:
                schemas_by_name[schema.name] = schema

        async def refresh_tools():
            toolkits_by_name.clear()

            visible_tools = await self._room.agents.list_toolkits(
                participant_id=participant_id
            )
            for toolkit_description in visible_tools:
                toolkits_by_name[toolkit_description.name] = toolkit_description

        installed = False

        await refresh_tools()
        await refresh_schemas()

        builtin_agents_url = "http://localhost:8080"

        for requirement in self.get_requirements():
            if isinstance(requirement, RequiredToolkit):
                if requirement.name == "ui":
                    # TODO: maybe requirements can be marked as non installable?
                    continue

                if requirement.name not in toolkits_by_name:
                    if not requirement.callable:
                        if requirement.timeout == 0:
                            logger.info(
                                f"{self.name} not waiting for toolkit {requirement.name}"
                            )
                            continue

                        async with asyncio.timeout(requirement.timeout):
                            logger.info(
                                f"{self.name} waiting for toolkit {requirement.name}"
                            )

                            while requirement.name not in toolkits_by_name:
                                await refresh_tools()
                                await asyncio.sleep(1)

                    else:
                        installed = True
                        logger.info(
                            f"{self.name} calling required tool into room {requirement.name}"
                        )

                        if requirement.name.startswith(
                            "https://"
                        ) or requirement.name.startswith("http://"):
                            url = requirement.name
                        else:
                            url = f"{builtin_agents_url}/toolkits/{requirement.name}"

                        await self._room.agents.make_call(
                            url=url, name=requirement.name, arguments={}
                        )

            elif isinstance(requirement, RequiredSchema):
                if requirement.schema is not None:
                    logger.info(
                        f"{self.name} installing required schema {requirement.name} from json"
                    )
                    handle = await self._room.storage.open(
                        path=f".schemas/{requirement.name}.json", overwrite=True
                    )
                    await self._room.storage.write(
                        handle=handle,
                        data=json.dumps(requirement.schema.to_json()).encode(),
                    )
                    await self._room.storage.close(handle=handle)

                elif requirement.name not in schemas_by_name:
                    installed = True

                    if not requirement.callable:
                        if requirement.timeout == 0:
                            logger.info(
                                f"{self.name} not waiting for schema {requirement.name}"
                            )
                            continue

                        async with asyncio.timeout(requirement.timeout):
                            logger.info(
                                f"{self.name} waiting for schema {requirement.name}"
                            )

                            while requirement.name not in schemas_by_name:
                                await refresh_schemas()
                                await asyncio.sleep(1)

                    else:
                        logger.info(
                            f"{self.name} installing required schema {requirement.name} from registry"
                        )

                        if requirement.name.startswith(
                            "https://"
                        ) or requirement.name.startswith("http://"):
                            url = requirement.name
                        else:
                            url = f"{builtin_agents_url}/schemas/{requirement.name}"

                        await self._room.agents.make_call(
                            url=url, name=requirement.name, arguments={}
                        )

            elif isinstance(requirement, RequiredTable):
                logger.info(
                    f"ensuring required table exists {requirement.name} in {requirement.namespace}"
                )

                await install_required_table(room=self.room, table=requirement)

            else:
                raise RoomException("unsupported requirement")

        if installed:
            await asyncio.sleep(5)

    async def get_toolkits(
        self, context: ToolContext, remote_toolkits: list[RequiredToolkit]
    ):
        tool_target = context.caller
        if context.on_behalf_of is not None:
            tool_target = context.on_behalf_of

        toolkits_by_name = dict[str, ToolkitDescription]()

        toolkits = list[Toolkit]()

        visible_tools = await self._room.agents.list_toolkits(
            participant_id=tool_target.id
        )

        for toolkit_description in visible_tools:
            toolkits_by_name[toolkit_description.name] = toolkit_description

        for required_toolkit in remote_toolkits:
            if isinstance(required_toolkit, RequiredToolkit):
                toolkit = toolkits_by_name.get(required_toolkit.name, None)
                if toolkit is None:
                    if context.on_behalf_of is not None:
                        raise RoomException(
                            f"unable to get toolkit {required_toolkit.name} on behalf of {context.on_behalf_of}"
                        )
                    else:
                        raise RoomException(
                            f"unable to get toolkit {required_toolkit.name} for caller {context.caller.id}"
                        )

                room_tools = list[RoomTool]()

                if required_toolkit.tools is None:
                    for tool_description in toolkit.tools:
                        tool = RoomTool(
                            on_behalf_of_id=tool_target.id,
                            toolkit_name=toolkit.name,
                            name=tool_description.name,
                            description=tool_description.description,
                            input_schema=tool_description.input_schema,
                            title=tool_description.title,
                            thumbnail_url=tool_description.thumbnail_url,
                            participant_id=tool_target.id,
                            defs=tool_description.defs,
                        )
                        room_tools.append(tool)

                else:
                    tools_by_name = dict[str, ToolDescription]()
                    for tool_description in toolkit.tools:
                        tools_by_name[tool_description.name] = tool_description

                    for required_tool in required_toolkit.tools:
                        tool_description = tools_by_name.get(required_tool, None)
                        if tool_description is None:
                            raise RoomException(
                                f"unable to locate required tool {required_tool} in toolkit {required_toolkit.name}"
                            )

                        tool = RoomTool(
                            on_behalf_of_id=tool_target.id,
                            toolkit_name=toolkit.name,
                            name=tool_description.name,
                            description=tool_description.description,
                            input_schema=tool_description.input_schema,
                            title=tool_description.title,
                            thumbnail_url=tool_description.thumbnail_url,
                            participant_id=tool_target.id,
                            defs=tool_description.defs,
                        )
                        room_tools.append(tool)

                toolkits.append(
                    Toolkit(
                        name=toolkit.name,
                        title=toolkit.title,
                        description=toolkit.description,
                        thumbnail_url=toolkit.thumbnail_url,
                        tools=room_tools,
                    )
                )

        return toolkits

    async def get_required_toolkits(self, context: ToolContext) -> list[Toolkit]:
        return await self.get_toolkits(context, self.requires)


class TaskRunner(SingleRoomAgent):
    def __init__(
        self,
        *,
        name,
        title=None,
        description=None,
        requires=None,
        supports_tools: Optional[bool] = None,
        input_schema: dict,
        output_schema: Optional[dict] = None,
        labels: Optional[list[str]] = None,
        toolkits: Optional[list[Toolkit]] = None,
        annotations: Optional[list[str]] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            labels=labels,
        )

        if toolkits is None:
            toolkits = []

        self._toolkits = toolkits

        self._registration_id = None

        if input_schema is None:
            input_schema = no_arguments_schema(
                description="execute the agent",
            )

        if supports_tools is None:
            supports_tools = False

        self._supports_tools = supports_tools
        self._input_schema = input_schema
        self._output_schema = output_schema
        self._annotations = annotations

    async def validate_arguments(self, arguments: dict):
        validate(arguments, self.input_schema)

    async def validate_response(self, response: dict):
        if self.output_schema is not None:
            validate(response, self.output_schema)

    async def ask(
        self,
        *,
        context: AgentCallContext,
        arguments: dict,
        attachment: Optional[bytes] = None,
    ) -> dict:
        raise Exception("Not implemented")

    @property
    def supports_tools(self):
        return self._supports_tools

    @property
    def annotations(self):
        return self._annotations

    @property
    def input_schema(self):
        return self._input_schema

    @property
    def output_schema(self):
        return self._output_schema

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "input_schema": self.input_schema,
            "output_schema": self.output_schema,
            "requires": list(map(lambda x: x.to_json(), self.requires)),
            "supports_tools": self.supports_tools,
            "labels": self.labels,
            "annotations": self.annotations,
        }

    async def _register(self):
        self._registration_id = (
            await self._room.send_request(
                "agent.register_agent",
                {
                    "name": self.name,
                    "title": self.title,
                    "description": self.description,
                    "input_schema": self.input_schema,
                    "output_schema": self.output_schema,
                    "requires": list(map(lambda x: x.to_json(), self.requires)),
                    "supports_tools": self.supports_tools,
                    "labels": self.labels,
                    "annotations": self.annotations,
                },
            )
        )["id"]

    async def _unregister(self):
        await self._room.send_request(
            "agent.unregister_agent", {"id": self._registration_id}
        )
        self._registration_id = None

    async def start(self, *, room: RoomClient):
        await super().start(room=room)

        self._room.protocol.register_handler("agent.ask", self._ask)
        await self._register()

    async def stop(self):
        if self.room.protocol.is_open:
            await self._unregister()
        else:
            logger.info(
                f"disconnected '{self.name}' from room, this will automatically happen when all the users leave the room. agents will not keep the room open"
            )

        self._room.protocol.unregister_handler("agent.ask", self._ask)

        await super().stop()

    async def _ask(
        self, protocol: Protocol, message_id: int, msg_type: str, data: bytes
    ):
        async def worker():
            # Decode and parse the message
            message, attachment = unpack_message(data)
            logger.info("agent got message %s", message)
            args = message["arguments"]
            task_id = message["task_id"]
            toolkits_json = message["toolkits"]

            # context_json = message["context"]

            chat_context = None

            try:
                chat_context = await self.init_chat_context()

                caller: Participant | None = None
                on_behalf_of: Participant | None = None
                on_behalf_of_id = message.get("on_behalf_of_id", None)

                for participant in self._room.messaging.get_participants():
                    if message["caller_id"] == participant.id:
                        caller = participant
                        break

                    if on_behalf_of_id == participant.id:
                        on_behalf_of = participant
                        break

                if caller is None:
                    caller = RemoteParticipant(
                        id=message["caller_id"], role="user", attributes={}
                    )

                if on_behalf_of_id is not None and on_behalf_of is None:
                    on_behalf_of = RemoteParticipant(
                        id=message["on_behalf_of_id"], role="user", attributes={}
                    )

                tool_target = caller
                if on_behalf_of is not None:
                    tool_target = on_behalf_of

                toolkits = [
                    *self._toolkits,
                    *await self.get_required_toolkits(
                        context=ToolContext(
                            room=self.room,
                            caller=caller,
                            on_behalf_of=on_behalf_of,
                            caller_context={"chat": chat_context.to_json()},
                        )
                    ),
                ]

                context = AgentCallContext(
                    chat=chat_context,
                    room=self.room,
                    caller=caller,
                    on_behalf_of=on_behalf_of,
                    toolkits=toolkits,
                )

                for toolkit_json in toolkits_json:
                    tools = []
                    for tool_json in toolkit_json["tools"]:
                        tools.append(
                            RoomTool(
                                on_behalf_of_id=on_behalf_of_id,
                                participant_id=tool_target.id,
                                toolkit_name=toolkit_json["name"],
                                name=tool_json["name"],
                                title=tool_json["title"],
                                description=tool_json["description"],
                                input_schema=tool_json["input_schema"],
                                thumbnail_url=toolkit_json["thumbnail_url"],
                                defs=tool_json.get("defs", None),
                            )
                        )

                    context.toolkits.append(
                        Toolkit(
                            name=toolkit_json["name"],
                            title=toolkit_json["title"],
                            description=toolkit_json["description"],
                            thumbnail_url=toolkit_json["thumbnail_url"],
                            tools=tools,
                        )
                    )

                if attachment is not None and len(attachment) > 0:
                    response = await self.ask(
                        context=context, arguments=args, attachment=attachment
                    )
                else:
                    response = await self.ask(context=context, arguments=args)

                await protocol.send(
                    type="agent.ask_response",
                    data=pack_message(
                        {
                            "task_id": task_id,
                            "answer": response,
                            "caller_context": chat_context.to_json(),
                        }
                    ),
                )

            except Exception as e:
                logger.error("Task runner failed to complete task", exc_info=e)
                if chat_context is not None:
                    await protocol.send(
                        type="agent.ask_response",
                        data=pack_message(
                            {
                                "task_id": task_id,
                                "error": str(e),
                                "caller_context": chat_context.to_json(),
                            }
                        ),
                    )
                else:
                    await protocol.send(
                        type="agent.ask_response",
                        data=pack_message(
                            {
                                "task_id": task_id,
                                "error": str(e),
                            }
                        ),
                    )

        def on_done(task: asyncio.Task):
            task.result()

        task = asyncio.create_task(worker())
        task.add_done_callback(on_done)


class RunTaskTool(Tool):
    def __init__(
        self,
        *,
        name: str,
        agent_name: str,
        input_schema: dict,
        rules=None,
        thumbnail_url=None,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            input_schema=input_schema,
            rules=rules,
            thumbnail_url=thumbnail_url,
            title=title,
            description=description,
        )

        self._agent_name = agent_name

    async def execute(self, context: ToolContext, **kwargs):
        return await context.room.agents.ask(agent=self._agent_name, arguments=kwargs)


async def make_run_task_tool(context: ToolContext, toolkit: RequiredToolkit):
    agents = await context.room.agents.list_agents()
    tools = []
    for agent_name in toolkit.tools:
        agent = next((x for x in agents if x.name == agent_name), None)

        if agent is None:
            raise RoomException(f"agent was not found in the room {agent_name}")

        tools.append(
            RunTaskTool(
                agent_name=agent_name,
                name=f"run_{agent_name}",
                input_schema=agent.input_schema,
                title=agent.title,
                description=agent.description,
            )
        )

    return Toolkit(name="agents", tools=tools)



from meshagent.agents.agent import SingleRoomAgent, AgentChatContext
from meshagent.api.chan import Chan
from meshagent.api import (
    RoomMessage,
    RoomClient,
    RemoteParticipant,
    Participant,
    Requirement,
    Element,
    MeshDocument,
    RequiredSchema,
)
from meshagent.tools import Toolkit, ToolContext, make_toolkits, ToolkitBuilder
from meshagent.agents.adapter import LLMAdapter, ToolResponseAdapter
from meshagent.openai.tools.responses_adapter import (
    ReasoningTool,
)
import asyncio
from typing import Optional, Callable
import logging
import uuid
import datetime
import base64
from asyncio import CancelledError
from meshagent.api import RoomException
from pydantic import BaseModel
from opentelemetry import trace
import shlex
import json

from pathlib import Path
from meshagent.agents.skills import to_prompt

from meshagent.agents.thread_schema import thread_schema

tracer = trace.get_tracer("meshagent.chatbot")

logger = logging.getLogger("chat")


class ChatBotReasoningTool(ReasoningTool):
    def __init__(self, *, room: RoomClient, thread_context: "ChatThreadContext"):
        super().__init__()
        self.thread_context = thread_context
        self.room = room

        self._reasoning_element = None
        self._reasoning_item = None

    def _get_messages_element(self):
        messages = self.thread_context.thread.root.get_children_by_tag_name("messages")
        if len(messages) > 0:
            return messages[0]
        return None

    async def on_reasoning_summary_part_added(
        self,
        context: ToolContext,
        *,
        item_id: str,
        output_index: int,
        part: dict,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        el = self._get_messages_element()
        if el is None:
            logger.warning("missing messages element, cannot log reasoning")
        else:
            self._reasoning_element = el.append_child("reasoning", {"summary": ""})

    async def on_reasoning_summary_part_done(
        self,
        context: ToolContext,
        *,
        item_id: str,
        output_index: int,
        part: dict,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        self._reasoning_element = None

    async def on_reasoning_summary_text_delta(
        self,
        context: ToolContext,
        *,
        delta: str,
        output_index: int,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        el = self._reasoning_element
        el.set_attribute("summary", el.get_attribute("summary") + delta)

    async def on_reasoning_summary_text_done(
        self,
        context: ToolContext,
        *,
        item_id: str,
        output_index: int,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        pass


def get_online_participants(
    *,
    room: RoomClient,
    thread: MeshDocument,
    exclude: Optional[list[Participant]] = None,
) -> list[RemoteParticipant]:
    results = list[RemoteParticipant]()

    for prop in thread.root.get_children():
        if prop.tag_name == "members":
            for member in prop.get_children():
                for online in room.messaging.get_participants():
                    if online.get_attribute("name") == member.get_attribute("name"):
                        if exclude is None or online not in exclude:
                            results.append(online)

    return results


class ChatThreadContext:
    def __init__(
        self,
        *,
        chat: AgentChatContext,
        thread: MeshDocument,
        path: str,
        participants: Optional[list[RemoteParticipant]] = None,
        event_handler: Optional[Callable[[dict], None]] = None,
    ):
        self.thread = thread
        if participants is None:
            participants = []

        self.participants = participants
        self.chat = chat
        self.path = path
        self._event_handler = event_handler

    def emit(self, event: dict):
        if self._event_handler is not None:
            self._event_handler(event)


class ChatBot(SingleRoomAgent):
    def __init__(
        self,
        *,
        name,
        title=None,
        description=None,
        requires: Optional[list[Requirement]] = None,
        llm_adapter: LLMAdapter,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        client_rules: Optional[dict[str, list[str]]] = None,
        auto_greet_message: Optional[str] = None,
        empty_state_title: Optional[str] = None,
        labels: Optional[list[str]] = None,
        decision_model: Optional[str] = None,
        always_reply: Optional[bool] = None,
        skill_dirs: Optional[list[str]] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            labels=labels,
        )

        if toolkits is None:
            toolkits = []

        self._decision_model = (
            "gpt-4.1-mini" if decision_model is None else decision_model
        )

        if always_reply is None:
            always_reply = False

        self._always_reply = always_reply

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter

        self._message_channels = dict[str, Chan[RoomMessage]]()

        self._room: RoomClient | None = None
        self._toolkits = toolkits
        self._client_rules = client_rules

        if rules is None:
            rules = []

        self._rules = rules
        self._is_typing = dict[str, asyncio.Task]()
        self._auto_greet_message = auto_greet_message

        if empty_state_title is None:
            empty_state_title = "How can I help you?"
        self._empty_state_title = empty_state_title

        self._thread_tasks = dict[str, asyncio.Task]()
        self._open_threads = {}

        self._skill_dirs = skill_dirs

    async def _send_and_save_chat(
        self,
        thread: MeshDocument,
        path: str,
        to: RemoteParticipant,
        id: str,
        text: str,
        thread_attributes: dict,
    ):
        messages = None

        for prop in thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop
                break

        if messages is None:
            raise RoomException("messages element was not found in thread document")

        with tracer.start_as_current_span("chatbot.thread.message") as span:
            span.set_attributes(thread_attributes)
            span.set_attribute("role", "assistant")
            span.set_attribute(
                "from_participant_name",
                self.room.local_participant.get_attribute("name"),
            )
            span.set_attributes({"id": id, "text": text})

            await self.room.messaging.send_message(
                to=to, type="chat", message={"path": path, "text": text}
            )

            messages.append_child(
                tag_name="message",
                attributes={
                    "id": id,
                    "text": text,
                    "created_at": datetime.datetime.now(datetime.timezone.utc)
                    .isoformat()
                    .replace("+00:00", "Z"),
                    "author_name": self.room.local_participant.get_attribute("name"),
                },
            )

    async def _greet(
        self,
        *,
        thread: MeshDocument,
        path: str,
        chat_context: AgentChatContext,
        participant: RemoteParticipant,
        thread_attributes: dict,
    ):
        if self._auto_greet_message is not None:
            chat_context.append_user_message(self._auto_greet_message)
            await self._send_and_save_chat(
                id=str(uuid.uuid4()),
                to=RemoteParticipant(id=participant.id),
                thread=thread,
                path=path,
                text=self._auto_greet_message,
                thread_attributes=thread_attributes,
            )

    def get_requirements(self):
        return [
            *super().get_requirements(),
            RequiredSchema(name="thread", schema=thread_schema),
        ]

    async def get_online_participants(
        self, *, thread: MeshDocument, exclude: Optional[list[Participant]] = None
    ):
        return get_online_participants(room=self._room, thread=thread, exclude=exclude)

    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return []

    async def get_thread_toolkits(
        self, *, thread_context: ChatThreadContext, participant: RemoteParticipant
    ) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=self.room.local_participant,
                on_behalf_of=participant,
                caller_context={"chat": thread_context.chat.to_json()},
                event_handler=thread_context.emit,
            )
        )

        toolkits.append(
            Toolkit(
                name="reasoning",
                tools=[
                    ChatBotReasoningTool(
                        room=self._room,
                        thread_context=thread_context,
                    )
                ],
            )
        )

        return [*self._toolkits, *toolkits]

    async def init_chat_context(self) -> AgentChatContext:
        context = self._llm_adapter.create_chat_context()
        context.append_rules(self._rules)
        return context

    async def open_thread(self, *, path: str):
        logger.info(f"opening thread {path}")
        if path not in self._open_threads:
            fut = asyncio.ensure_future(
                self.room.sync.open(path=path, schema=thread_schema)
            )
            self._open_threads[path] = fut

        return await self._open_threads[path]

    async def close_thread(self, *, path: str):
        logger.info(f"closing thread {path}")

        if path in self._open_threads:
            del self._open_threads[path]

        return await self.room.sync.close(path=path)

    async def load_thread_context(self, *, thread_context: ChatThreadContext):
        """
        load the thread from the thread document by inserting the current messages in the thread into the chat context
        """
        thread = thread_context.thread
        chat_context = thread_context.chat
        for prop in thread.root.get_children():
            if prop.tag_name == "messages":
                doc_messages = prop

                for element in doc_messages.get_children():
                    if isinstance(element, Element) and element.tag_name == "message":
                        msg = element["text"]
                        if element[
                            "author_name"
                        ] == self.room.local_participant.get_attribute("name"):
                            chat_context.append_assistant_message(msg)
                        else:
                            chat_context.append_user_message(
                                self.format_message(
                                    user_name=element["author_name"],
                                    message=msg,
                                    iso_timestamp=element["created_at"],
                                )
                            )

                        for child in element.get_children():
                            if child.tag_name == "file":
                                chat_context.append_assistant_message(
                                    f"the user attached a file with the path '{child.get_attribute('path')}'"
                                )

                break

        if doc_messages is None:
            raise Exception("thread was not properly initialized")

    async def prepare_llm_context(self, *, thread_context: ChatThreadContext):
        """
        called prior to sending the request to the LLM in case the agent needs to modify the context prior to sending
        """
        pass

    async def _process_llm_events(
        self,
        *,
        thread_context: ChatThreadContext,
        llm_messages: asyncio.Queue,
        thread_attributes: dict,
    ):
        thread = thread_context.thread
        doc_messages = None
        for prop in thread.root.get_children():
            if prop.tag_name == "messages":
                doc_messages = prop
                break

        if doc_messages is None:
            raise RoomException("messages element is missing from thread document")

        context_message = None
        updates = asyncio.Queue()

        # throttle updates so we don't send too many syncs over the wire at once
        async def update_thread():
            try:
                changes = {}
                while True:
                    try:
                        element, partial = updates.get_nowait()
                        changes[element] = partial

                    except asyncio.QueueEmpty:
                        for e, p in changes.items():
                            e["text"] = p

                        changes.clear()

                        e, p = await updates.get()
                        changes[e] = p

                        await asyncio.sleep(0.1)

            except asyncio.QueueShutDown:
                # flush any pending changes
                for e, p in changes.items():
                    e["text"] = p

                changes.clear()
                pass

        update_thread_task = asyncio.create_task(update_thread())
        try:
            while True:
                evt = await llm_messages.get()

                if evt["type"] == "response.content_part.added":
                    partial = ""

                    content_element = doc_messages.append_child(
                        tag_name="message",
                        attributes={
                            "text": "",
                            "created_at": datetime.datetime.now(datetime.timezone.utc)
                            .isoformat()
                            .replace("+00:00", "Z"),
                            "author_name": self.room.local_participant.get_attribute(
                                "name"
                            ),
                        },
                    )

                    context_message = {"role": "assistant", "content": ""}
                    thread_context.chat.messages.append(context_message)

                elif evt["type"] == "response.output_text.delta":
                    partial += evt["delta"]
                    updates.put_nowait((content_element, partial))
                    context_message["content"] = partial

                elif evt["type"] == "response.output_text.done":
                    content_element = None

                    with tracer.start_as_current_span("chatbot.thread.message") as span:
                        span.set_attribute(
                            "from_participant_name",
                            self.room.local_participant.get_attribute("name"),
                        )
                        span.set_attribute("role", "assistant")
                        span.set_attributes(thread_attributes)
                        span.set_attributes({"text": evt["text"]})

                        for participant in get_online_participants(
                            room=self._room, thread=thread_context.thread
                        ):
                            if participant.id != self._room.local_participant.id:
                                logger.info(
                                    f"replying to {participant.get_attribute('name')}"
                                )
                                self._room.messaging.send_message_nowait(
                                    to=participant,
                                    type="chat",
                                    message={
                                        "type": "chat",
                                        "path": thread_context.path,
                                        "text": evt["text"],
                                    },
                                )
                elif evt["type"] == "response.image_generation_call.partial_image":
                    await self.handle_image_generation_partial(
                        thread_context=thread_context,
                        llm_messages=llm_messages,
                        event=evt,
                    )

                elif evt["type"] == "meshagent.handler.added":
                    item = evt["item"]
                    print(f"{item}", flush=True)
                    if item["type"] == "shell_call":
                        await self.handle_shell_call_output(
                            thread_context=thread_context,
                            llm_messages=llm_messages,
                            item=item,
                        )

                    elif item["type"] == "local_shell_call":
                        await self.handle_local_shell_call_output(
                            thread_context=thread_context,
                            llm_messages=llm_messages,
                            item=item,
                        )

        except asyncio.QueueShutDown:
            pass
        finally:
            updates.shutdown()

        await update_thread_task

    async def handle_image_generation_partial(
        self,
        *,
        thread_context: ChatThreadContext,
        llm_messages: asyncio.Queue,
        event: dict,
    ):
        item_id = event["item_id"]
        partial_image_b64 = event["partial_image_b64"]
        output_format = event["output_format"]

        messages = thread_context.thread.root.get_children_by_tag_name("messages")[0]

        if output_format is None:
            output_format = "png"

        image_name = f"{str(uuid.uuid4())}.{output_format}"

        handle = await self.room.storage.open(path=image_name)
        await self.room.storage.write(
            handle=handle, data=base64.b64decode(partial_image_b64)
        )
        await self.room.storage.close(handle=handle)

        messages = None

        logger.info(f"A partial was saved at the path {image_name}")

        for prop in thread_context.thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop
                break

        for child in messages.get_children():
            if child.get_attribute("id") == item_id:
                for file in child.get_children():
                    file.set_attribute("path", image_name)

                return

        message_element = messages.append_child(
            tag_name="message",
            attributes={
                "id": item_id,
                "text": "",
                "created_at": datetime.datetime.now(datetime.timezone.utc)
                .isoformat()
                .replace("+00:00", "Z"),
                "author_name": self.room.local_participant.get_attribute("name"),
            },
        )
        message_element.append_child(tag_name="file", attributes={"path": image_name})

    async def handle_local_shell_call_output(
        self,
        *,
        thread_context: ChatThreadContext,
        llm_messages: asyncio.Queue,
        item: dict,
    ):
        messages = thread_context.thread.root.get_children_by_tag_name("messages")[0]

        action = item["action"]
        command = action["command"]
        working_directory = action["working_directory"]

        for prop in thread_context.thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop
                break

        exec_element = messages.append_child(
            tag_name="exec",
            attributes={"command": shlex.join(command), "pwd": working_directory},
        )

        evt = await llm_messages.get()

        if evt["type"] != "meshagent.handler.done":
            raise RoomException("expected meshagent.handler.done")

        error = evt.get("error")
        item = evt.get("item")

        if error is not None:
            pass

        if item is not None:
            if item["type"] != "local_shell_call_output":
                raise RoomException("expected local_shell_call_output")

            exec_element.set_attribute("result", item["output"])

    async def handle_shell_call_output(
        self,
        *,
        thread_context: ChatThreadContext,
        llm_messages: asyncio.Queue,
        item: dict,
    ):
        messages = thread_context.thread.root.get_children_by_tag_name("messages")[0]

        action = item["action"]
        commands = action["commands"]

        exec_elements = []
        for command in commands:
            exec_element = messages.append_child(
                tag_name="exec",
                attributes={"command": command},
            )
            exec_elements.append(exec_element)

        evt = await llm_messages.get()

        if evt["type"] != "meshagent.handler.done":
            raise RoomException("expected meshagent.handler.done")

        error = evt.get("error")
        item = evt.get("item")

        if error is not None:
            pass

        if item is not None:
            if item["type"] != "shell_call_output":
                raise RoomException("expected shell_call_output")

            results = item["output"]

            for i in range(0, len(results)):
                result = results[i]
                exec_element = exec_elements[i]
                if "exit_code" in result["outcome"]:
                    exec_element.set_attribute(
                        "exit_code", result["outcome"]["exit_code"]
                    )

                exec_element.set_attribute("outcome", result["outcome"]["type"])
                exec_element.set_attribute("stdout", result["stdout"])
                exec_element.set_attribute("stderr", result["stderr"])

    def get_thread_members(self, *, thread: MeshDocument) -> list[str]:
        results = []

        for prop in thread.root.get_children():
            if prop.tag_name == "members":
                for member in prop.get_children():
                    results.append(member.get_attribute("name"))

        return results

    async def should_reply(
        self,
        *,
        context: ChatThreadContext,
        has_more_than_one_other_user: bool,
        toolkits: list[Toolkit],
        from_user: RemoteParticipant,
        online: list[Participant],
    ):
        if not has_more_than_one_other_user or self._always_reply:
            return True

        online_set = {}

        all_members = []
        online_members = []

        for m in self.get_thread_members(thread=context.thread):
            all_members.append(m)

        for o in online:
            if o.get_attribute("name") not in online_set:
                online_set[o.get_attribute("name")] = True
                online_members.append(o.get_attribute("name"))

        logger.info(
            "multiple participants detected, checking whether agent should reply to conversation"
        )

        cloned_context = context.chat.copy()
        cloned_context.replace_rules(
            rules=[
                "examine the conversation so far and return whether the user is expecting a reply from you or another user as the next message in the conversation",
                f'your name (the assistant) is "{self.room.local_participant.get_attribute("name")}"',
                "if the user mentions a person with another name, they aren't talking to you unless they also mention you",
                "if the user poses a question to everyone, they are talking to you",
                f"members of thread are currently {all_members}",
                f"users online currently are {online_members}",
            ]
        )
        response = await self._llm_adapter.next(
            context=cloned_context,
            room=self._room,
            model=self._decision_model or self._llm_adapter.default_model(),
            on_behalf_of=from_user,
            toolkits=[],
            output_schema={
                "type": "object",
                "required": ["reasoning", "expecting_assistant_reply", "next_user"],
                "additionalProperties": False,
                "properties": {
                    "reasoning": {
                        "type": "string",
                        "description": "explain why you think the user was or was not expecting you to reply",
                    },
                    "next_user": {
                        "type": "string",
                        "description": "who would be expectd to send the next message in the conversation",
                    },
                    "expecting_assistant_reply": {"type": "boolean"},
                },
            },
        )

        logger.info(f"should reply check returned {response}")

        return response["expecting_assistant_reply"]

    async def handle_user_message(
        self,
        *,
        context: ChatThreadContext,
        toolkits: list[Toolkit],
        model: str,
        from_user: RemoteParticipant,
        event_handler,
    ):
        online = await self.get_online_participants(
            thread=context.thread, exclude=[self.room.local_participant]
        )

        for participant in get_online_participants(
            room=self._room, thread=context.thread
        ):
            self._room.messaging.send_message_nowait(
                to=participant,
                type="listening",
                message={"listening": True, "path": context.path},
            )

        has_more_than_one_other_user = False

        thread_participants = []

        for member_name in self.get_thread_members(thread=context.thread):
            thread_participants.append(member_name)
            if member_name != self._room.local_participant.get_attribute(
                "name"
            ) and member_name != from_user.get_attribute("name"):
                has_more_than_one_other_user = True
                break

        context.chat.metadata["thread_participants"] = thread_participants

        reply = await self.should_reply(
            has_more_than_one_other_user=has_more_than_one_other_user,
            online=online,
            context=context,
            toolkits=toolkits,
            from_user=from_user,
        )

        for participant in get_online_participants(
            room=self._room, thread=context.thread
        ):
            self._room.messaging.send_message_nowait(
                to=participant,
                type="listening",
                message={"listening": False, "path": context.path},
            )

        if not reply:
            return

        for participant in get_online_participants(
            room=self._room, thread=context.thread
        ):
            self._room.messaging.send_message_nowait(
                to=participant,
                type="thinking",
                message={"thinking": True, "path": context.path},
            )

        self.prepare_chat_context(chat_context=context.chat)

        await self._llm_adapter.next(
            context=context.chat,
            room=self._room,
            toolkits=toolkits,
            tool_adapter=self._tool_adapter,
            event_handler=event_handler,
            model=model,
            on_behalf_of=from_user,
        )

    async def get_rules(
        self, *, thread_context: ChatThreadContext, participant: RemoteParticipant
    ):
        rules = [*self._rules]

        if self._skill_dirs is not None and len(self._skill_dirs) > 0:
            rules.append(
                "You have access to to following skills which follow the agentskills spec:"
            )
            rules.append(await to_prompt([*(Path(p) for p in self._skill_dirs)]))
            rules.append(
                "Use the shell tool to find out more about skills and execute them when they are required"
            )

        client = participant.get_attribute("client")

        if self._client_rules is not None and client is not None:
            cr = self._client_rules.get(client)
            if cr is not None:
                rules.extend(cr)

        # Without this rule 5.2 / 5.1 like to start their messages with things like "I could say"
        rules.append("based on the previous transcript, take your turn and respond")

        return rules

    async def on_chat_received(
        self,
        *,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict,
    ):
        rules = await self.get_rules(
            thread_context=thread_context, participant=from_participant
        )
        thread_context.chat.replace_rules(rules)

    def format_message(self, *, user_name: str, message: str, iso_timestamp: str):
        return f"{user_name} said at {iso_timestamp}: {message}"

    def prepare_chat_context(self, *, chat_context: AgentChatContext):
        pass

    async def _spawn_thread(self, path: str, messages: Chan[RoomMessage]):
        logger.debug("chatbot is starting a thread", extra={"path": path})
        chat_context = await self.init_chat_context()
        opened = False

        current_file = None
        thread_context = None

        thread_attributes = None

        thread = None

        llm_messages = asyncio.Queue[dict]()
        llm_task = None

        def handle_event(evt):
            if isinstance(evt, BaseModel):
                evt = evt.model_dump(mode="json")
            llm_messages.put_nowait(evt)

        try:
            received = None

            while True:
                logger.debug(f"waiting for message on thread {path}")
                received = await messages.recv()
                logger.debug(f"received message on thread {path}: {received.type}")

                chat_with_participant = None
                for participant in self._room.messaging.get_participants():
                    if participant.id == received.from_participant_id:
                        chat_with_participant = participant
                        break

                if chat_with_participant is None:
                    logger.warning(
                        "participant does not have messaging enabled, skipping message"
                    )
                    continue

                thread_attributes = {
                    "agent_name": self.name,
                    "agent_participant_id": self.room.local_participant.id,
                    "agent_participant_name": self.room.local_participant.get_attribute(
                        "name"
                    ),
                    "remote_participant_id": chat_with_participant.id,
                    "remote_participant_name": chat_with_participant.get_attribute(
                        "name"
                    ),
                    "path": path,
                }

                if current_file != chat_with_participant.get_attribute("current_file"):
                    logger.info(
                        f"{chat_with_participant.get_attribute('name')} is now looking at {chat_with_participant.get_attribute('current_file')}"
                    )
                    current_file = chat_with_participant.get_attribute("current_file")

                if current_file is not None:
                    chat_context.append_assistant_message(
                        message=f"{chat_with_participant.get_attribute('name')} is currently viewing the file at the path: {current_file}"
                    )

                elif current_file is not None:
                    chat_context.append_assistant_message(
                        message=f"{chat_with_participant.get_attribute('name')} is not current viewing any files"
                    )

                if thread is None:
                    with tracer.start_as_current_span("chatbot.thread.open") as span:
                        span.set_attributes(thread_attributes)

                        thread = await self.open_thread(path=path)

                        thread_context = ChatThreadContext(
                            path=path,
                            chat=chat_context,
                            thread=thread,
                            participants=get_online_participants(
                                room=self.room, thread=thread
                            ),
                            event_handler=handle_event,
                        )

                        llm_task = asyncio.create_task(
                            self._process_llm_events(
                                thread_context=thread_context,
                                llm_messages=llm_messages,
                                thread_attributes=thread_attributes,
                            )
                        )

                        await self.load_thread_context(thread_context=thread_context)

                if received.type == "opened":
                    if not opened:
                        opened = True

                        await self._greet(
                            path=path,
                            chat_context=chat_context,
                            participant=chat_with_participant,
                            thread=thread,
                            thread_attributes=thread_attributes,
                        )
                elif received.type == "clear":
                    chat_context = await self.init_chat_context()
                    thread_context.chat = chat_context
                    messages_element: Element = thread.root.get_children_by_tag_name(
                        "messages"
                    )[0]
                    for child in list(messages_element.get_children()):
                        child.delete()

                elif received.type == "chat":
                    if thread is None:
                        logger.info("thread is not open", extra={"path": path})
                        break

                    logger.debug(
                        "chatbot received a chat",
                        extra={
                            "context": chat_context.id,
                            "participant_id": self.room.local_participant.id,
                            "participant_name": self.room.local_participant.get_attribute(
                                "name"
                            ),
                            "text": received.message["text"],
                        },
                    )

                    attachments = received.message.get("attachments", [])
                    text = received.message["text"]

                    for attachment in attachments:
                        chat_context.append_assistant_message(
                            message=f"the user attached a file at the path '{attachment['path']}'"
                        )

                    iso_timestamp = (
                        datetime.datetime.now(datetime.timezone.utc)
                        .isoformat()
                        .replace("+00:00", "Z")
                    )

                    chat_context.append_user_message(
                        message=self.format_message(
                            user_name=chat_with_participant.get_attribute("name"),
                            message=text,
                            iso_timestamp=iso_timestamp,
                        )
                    )

                if received is not None and received.type == "chat":
                    with tracer.start_as_current_span("chatbot.thread.message") as span:
                        span.set_attributes(thread_attributes)
                        span.set_attribute("role", "user")
                        span.set_attribute(
                            "from_participant_name",
                            chat_with_participant.get_attribute("name"),
                        )

                        attachments = received.message.get("attachments", [])
                        span.set_attribute("attachments", json.dumps(attachments))

                        text = received.message["text"]
                        span.set_attributes({"text": text})

                        try:
                            if thread_context is None:
                                thread_context = ChatThreadContext(
                                    path=path,
                                    chat=chat_context,
                                    thread=thread,
                                    participants=get_online_participants(
                                        room=self.room, thread=thread
                                    ),
                                )
                            else:
                                thread_context.participants = get_online_participants(
                                    room=self.room, thread=thread
                                )

                            await self.on_chat_received(
                                thread_context=thread_context,
                                from_participant=chat_with_participant,
                                message=received.message,
                            )

                            with tracer.start_as_current_span("chatbot.llm") as span:
                                try:
                                    with tracer.start_as_current_span(
                                        "get_thread_toolkits"
                                    ) as span:
                                        thread_toolkits = (
                                            await self.get_thread_toolkits(
                                                thread_context=thread_context,
                                                participant=chat_with_participant,
                                            )
                                        )

                                    with tracer.start_as_current_span(
                                        "get_thread_toolkit_builders"
                                    ) as span:
                                        thread_tool_providers = (
                                            self.get_toolkit_builders()
                                        )

                                    await self.prepare_llm_context(
                                        thread_context=thread_context
                                    )

                                    message_toolkits = [*thread_toolkits]

                                    model = received.message.get(
                                        "model", self._llm_adapter.default_model()
                                    )

                                    message_tools = received.message.get("tools")

                                    if (
                                        message_tools is not None
                                        and len(message_tools) > 0
                                    ):
                                        message_toolkits.extend(
                                            await make_toolkits(
                                                room=self.room,
                                                model=model,
                                                providers=thread_tool_providers,
                                                tools=message_tools,
                                            )
                                        )

                                    await self.handle_user_message(
                                        context=thread_context,
                                        toolkits=message_toolkits,
                                        event_handler=handle_event,
                                        model=model,
                                        from_user=chat_with_participant,
                                    )

                                except Exception as e:
                                    logger.error("An error was encountered", exc_info=e)
                                    await self._send_and_save_chat(
                                        thread=thread,
                                        to=chat_with_participant,
                                        path=path,
                                        id=str(uuid.uuid4()),
                                        text="There was an error while communicating with the LLM. Please try again later.",
                                        thread_attributes=thread_attributes,
                                    )

                        finally:

                            async def cleanup():
                                for participant in get_online_participants(
                                    room=self._room, thread=thread
                                ):
                                    self._room.messaging.send_message_nowait(
                                        to=participant,
                                        type="thinking",
                                        message={"thinking": False, "path": path},
                                    )

                            asyncio.shield(cleanup())

        finally:

            async def cleanup():
                llm_messages.shutdown()
                if llm_task is not None:
                    await llm_task

                if self.room is not None:
                    logger.info(f"thread was ended {path}")
                    logger.info("chatbot thread ended", extra={"path": path})

                    if thread is not None:
                        await self.close_thread(path=path)

            asyncio.shield(cleanup())

    def _get_message_channel(self, key: str) -> Chan[RoomMessage]:
        if key not in self._message_channels:
            chan = Chan[RoomMessage]()
            self._message_channels[key] = chan

        chan = self._message_channels[key]

        return chan

    async def stop(self):
        await super().stop()

        for thread in self._thread_tasks.values():
            thread.cancel()

        self._thread_tasks.clear()

    async def _on_get_thread_toolkits_message(self, *, message: RoomMessage):
        path = message.message["path"]

        thread_context = None
        if path in self._open_threads:
            thread = await self._open_threads[path]

            thread_context = ChatThreadContext(
                path=path,
                chat=AgentChatContext(),
                thread=thread,
                participants=get_online_participants(room=self.room, thread=thread),
            )

        if thread_context is None:
            logger.warning("thread toolkits requested for a thread that is not open")
            return

        chat_with_participant = None
        for participant in self._room.messaging.get_participants():
            if participant.id == message.from_participant_id:
                chat_with_participant = participant
                break

        if chat_with_participant is None:
            logger.warning(
                "participant does not have messaging enabled, skipping message"
            )
            return

        tool_providers = self.get_toolkit_builders()
        self._room.messaging.send_message_nowait(
            to=chat_with_participant,
            type="set_thread_tool_providers",
            message={
                "path": path,
                "tool_providers": [{"name": t.name} for t in tool_providers],
            },
        )

    async def start(self, *, room):
        await super().start(room=room)

        logger.debug("Starting chatbot")

        await self.room.local_participant.set_attribute(
            "empty_state_title", self._empty_state_title
        )

        def on_message(message: RoomMessage):
            if message.type == "get_thread_toolkit_builders":
                task = asyncio.create_task(
                    self._on_get_thread_toolkits_message(message=message)
                )

                def on_done(task: asyncio.Task):
                    try:
                        task.result()
                    except Exception as ex:
                        logger.error(f"unable to get tool providers {ex}", exc_info=ex)

                task.add_done_callback(on_done)

            if (
                message.type == "chat"
                or message.type == "opened"
                or message.type == "clear"
            ):
                path = message.message["path"]

                messages = self._get_message_channel(path)

                logger.debug(
                    f"queued incoming message for thread {path}: {message.type}"
                )

                messages.send_nowait(message)

                if path not in self._thread_tasks or self._thread_tasks[path].done():

                    def thread_done(task: asyncio.Task):
                        self._thread_tasks.pop(path)
                        self._message_channels.pop(path)
                        try:
                            task.result()
                        except CancelledError:
                            pass
                        except Exception as e:
                            logger.error(
                                f"The chat thread ended with an error {e}", exc_info=e
                            )

                    logger.debug(f"spawning chat thread for {path}")
                    task = asyncio.create_task(
                        self._spawn_thread(messages=messages, path=path)
                    )
                    task.add_done_callback(thread_done)

                    self._thread_tasks[path] = task

            elif message.type == "cancel":
                path = message.message["path"]
                if path in self._thread_tasks:
                    self._thread_tasks[path].cancel()

            elif message.type == "typing":

                def callback(task: asyncio.Task):
                    try:
                        task.result()
                    except CancelledError:
                        pass
                    except Exception:
                        pass

                async def remove_timeout(id: str):
                    await asyncio.sleep(1)
                    self._is_typing.pop(id)

                if message.from_participant_id in self._is_typing:
                    self._is_typing[message.from_participant_id].cancel()

                timeout = asyncio.create_task(
                    remove_timeout(id=message.from_participant_id)
                )
                timeout.add_done_callback(callback)

                self._is_typing[message.from_participant_id] = timeout

        room.messaging.on("message", on_message)

        if self._auto_greet_message is not None:

            def on_participant_added(participant: RemoteParticipant):
                # will spawn the initial thread
                self._get_message_channel(participant.id)

            room.messaging.on("participant_added", on_participant_added)

        logger.debug("Enabling chatbot messaging")
        await room.messaging.enable()



from typing import Optional

from jsonschema import validate, ValidationError
from meshagent.api.schema_util import prompt_schema, merge
from meshagent.api import Requirement
from meshagent.tools import Toolkit, make_toolkits, ToolkitBuilder
from meshagent.agents import TaskRunner
from meshagent.agents.agent import AgentCallContext
from meshagent.agents.adapter import LLMAdapter, ToolResponseAdapter
import tarfile
import io
import mimetypes


class LLMTaskRunner(TaskRunner):
    """
    A Task Runner that uses an LLM execution loop until the task is complete.
    """

    def __init__(
        self,
        *,
        name: str,
        llm_adapter: LLMAdapter,
        title: Optional[str] = None,
        description: Optional[str] = None,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        toolkits: Optional[list[Toolkit]] = None,
        requires: Optional[list[Requirement]] = None,
        supports_tools: bool = True,
        input_prompt: bool = True,
        input_schema: Optional[dict] = None,
        output_schema: Optional[dict] = None,
        rules: Optional[list[str]] = None,
        labels: Optional[list[str]] = None,
        annotations: Optional[list[str]] = None,
        client_rules: Optional[dict[str, list[str]]] = None,
    ):
        if input_schema is None:
            if input_prompt:
                input_schema = prompt_schema(
                    description="use a prompt to generate content"
                )

                toolkit_builders = self.get_toolkit_builders()
                if len(toolkit_builders) > 0:
                    toolkit_config_schemas = []

                    defs = None

                    for builder in toolkit_builders:
                        schema = builder.type.model_json_schema()
                        if schema.get("$defs") is not None:
                            if defs is None:
                                defs = {}

                            for k, v in schema["$defs"].items():
                                defs[k] = v

                        toolkit_config_schemas.append(schema)

                    input_schema = merge(
                        schema=input_schema,
                        additional_properties={
                            "tools": {
                                "type": "array",
                                "items": {
                                    "anyOf": toolkit_config_schemas,
                                },
                            },
                            "model": {"type": ["string", "null"]},
                        },
                    )

                    if defs is not None:
                        if input_schema.get("$defs") is None:
                            input_schema["$defs"] = {}

                        for k, v in defs.items():
                            input_schema["$defs"][k] = v

            else:
                input_schema = {
                    "type": "object",
                    "additionalProperties": False,
                    "required": [],
                    "properties": {},
                }

        static_toolkits = list(toolkits or [])

        super().__init__(
            name=name,
            title=title,
            description=description,
            input_schema=input_schema,
            output_schema=output_schema,
            requires=requires,
            supports_tools=supports_tools,
            labels=labels,
            toolkits=static_toolkits,
            annotations=annotations,
        )

        self._extra_rules = rules or []
        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter
        self.toolkits = static_toolkits
        self._client_rules = client_rules

    async def init_chat_context(self):
        chat = self._llm_adapter.create_chat_context()
        return chat

    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return []

    async def get_context_toolkits(self, *, context: AgentCallContext) -> list[Toolkit]:
        return []

    async def get_rules(self, *, context: AgentCallContext):
        rules = [*self._extra_rules]

        participant = context.caller
        client = participant.get_attribute("client")

        if self._client_rules is not None and client is not None:
            cr = self._client_rules.get(client)
            if cr is not None:
                rules.extend(cr)

        return rules

    async def ask(
        self,
        *,
        context: AgentCallContext,
        arguments: dict,
        attachment: Optional[bytes] = None,
    ):
        prompt = arguments.get("prompt")
        if prompt is None:
            raise ValueError("`prompt` is required")

        message_tools = arguments.get("tools")
        model = arguments.get("model", self._llm_adapter.default_model())

        context.chat.append_rules(await self.get_rules(context=context))

        context.chat.append_user_message(prompt)

        if attachment is not None:
            buf = io.BytesIO(attachment)
            with tarfile.open(fileobj=buf, mode="r:*") as tar:
                for member in tar.getmembers():
                    if member.isfile():
                        mime_type, encoding = mimetypes.guess_type(member.name)
                        f = tar.extractfile(member)
                        content = f.read()
                        if mime_type.startswith("image/"):
                            context.chat.append_image_message(
                                data=content, mime_type=mime_type
                            )
                        else:
                            context.chat.append_file_message(
                                filename=member.name, data=content, mime_type=mime_type
                            )

        combined_toolkits: list[Toolkit] = [
            *self.toolkits,
            *context.toolkits,
            *await self.get_context_toolkits(context=context),
        ]

        if message_tools is not None and len(message_tools) > 0:
            combined_toolkits.extend(
                await make_toolkits(
                    room=self.room,
                    model=model,
                    providers=self.get_toolkit_builders(),
                    tools=message_tools,
                )
            )

        resp = await self._llm_adapter.next(
            context=context.chat,
            room=context.room,
            toolkits=combined_toolkits,
            tool_adapter=self._tool_adapter,
            output_schema=self.output_schema,
        )

        # Validate the LLM output against the declared output schema if one was provided
        if self.output_schema:
            try:
                validate(instance=resp, schema=self.output_schema)
            except ValidationError as exc:
                raise RuntimeError("LLM output failed schema validation") from exc

        return resp


class DynamicLLMTaskRunner(LLMTaskRunner):
    """
    Same capabilities as LLMTaskRunner, but the caller supplies an arbitrary JSON-schema (`output_schema`) at runtime
    """

    def __init__(
        self,
        *,
        name: str,
        llm_adapter: LLMAdapter,
        supports_tools: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        annotations: Optional[list[str]] = None,
    ):
        input_schema = merge(
            schema=prompt_schema(description="use a prompt to generate content"),
            additional_properties={"output_schema": {"type": "object"}},
        )
        super().__init__(
            name=name,
            llm_adapter=llm_adapter,
            supports_tools=supports_tools,
            title=title,
            description=description,
            tool_adapter=tool_adapter,
            toolkits=toolkits,
            rules=rules,
            input_prompt=True,
            input_schema=input_schema,
            output_schema=None,
            annotations=annotations,
        )

    async def ask(self, *, context: AgentCallContext, arguments: dict):
        prompt = arguments.get("prompt")
        if prompt is None:
            raise ValueError("`prompt` is required")

        # Parse and pass JSON output schema provided at runtime
        output_schema_raw = arguments.get("output_schema")
        if output_schema_raw is None:
            raise ValueError("`output_schema` is required for DynamicLLMTaskRunner")

        # Make sure provided schema is a dict
        if not isinstance(output_schema_raw, dict):
            raise TypeError("`output_schema` must be a dict (JSON-schema object)")

        context.chat.append_user_message(prompt)

        combined_toolkits: list[Toolkit] = [*self.toolkits, *context.toolkits]

        resp = await self._llm_adapter.next(
            context=context.chat,
            room=context.room,
            toolkits=combined_toolkits,
            tool_adapter=self._tool_adapter,
            output_schema=output_schema_raw,
        )

        try:
            validate(instance=resp, schema=output_schema_raw)
        except ValidationError as exc:
            raise RuntimeError("LLM output failed caller schema validation") from exc

        return resp



from meshagent.api.schema import MeshSchema, ElementType, ChildProperty, ValueProperty

super_editor_document_schema = MeshSchema(
    root_tag_name="document",
    elements=[
        ElementType(
            tag_name="document",
            description="A SuperEditor document for use with Flutter apps",
            properties=[
                ChildProperty(
                    name="children",
                    child_tag_names=[
                        "block",
                        "horizontalRule",
                        "image",
                    ],
                ),
            ],
        ),
        ElementType(
            tag_name="block",
            description="Block element - this can be a paragraph, list-item or tasks",
            properties=[
                ValueProperty(name="blockType", type="string"),
                ValueProperty(name="textAlign", type="string"),
                ValueProperty(name="itemType", type="string"),
                ValueProperty(name="indent", type="number"),
                ChildProperty(name="children", child_tag_names=["text"]),
            ],
        ),
        ElementType(
            tag_name="image",
            description="Image element",
            properties=[
                ValueProperty(name="imageUrl", type="string"),
                ValueProperty(name="width", type="number"),
                ValueProperty(name="height", type="number"),
            ],
        ),
        ElementType(
            tag_name="horizontalRule",
            description="Horizontal rule element",
            properties=[],
        ),
        ElementType(
            tag_name="text",
            description="Text element",
            properties=[
                ValueProperty(name="text", type="string"),
            ],
        ),
    ],
)



from meshagent.agents.config import RulesConfig


def test_parse_empty_text():
    config = RulesConfig.parse(text="")

    assert config.rules == []
    assert config.client_rules == {}


def test_parse_global_rules_only():
    text = """\
first rule
second rule
third rule"""

    config = RulesConfig.parse(text=text)

    assert config.rules == ["first rule", "second rule", "third rule"]
    assert config.client_rules == {}


def test_comments():
    text = """\
# first rule
second rule
third rule"""

    config = RulesConfig.parse(text=text)

    assert config.rules == ["second rule", "third rule"]
    assert config.client_rules == {}


def test_parse_single_client_rules_only():
    text = """\
[web]
can browse the web
can use cookies
"""

    config = RulesConfig.parse(text=text)

    assert config.rules == []
    assert config.client_rules == {"web": ["can browse the web", "can use cookies"]}


def test_parse_global_and_client_rules():
    text = """\
global rule 1
global rule 2
[mobile]
use push notifications
limit data usage
"""

    config = RulesConfig.parse(text=text)

    assert config.rules == ["global rule 1", "global rule 2"]
    assert config.client_rules == {
        "mobile": ["use push notifications", "limit data usage"]
    }


def test_parse_multiple_clients():
    text = """\
[web]
show rich ui
support mouse
[mobile]
support touch
battery saver
"""

    config = RulesConfig.parse(text=text)

    assert config.rules == []
    assert config.client_rules == {
        "web": ["show rich ui", "support mouse"],
        "mobile": ["support touch", "battery saver"],
    }


def test_parse_remembers_current_client_until_next_header():
    text = """\
[client_a]
rule 1
rule 2
[client_b]
rule 3
"""

    config = RulesConfig.parse(text=text)

    assert config.client_rules == {
        "client_a": ["rule 1", "rule 2"],
        "client_b": ["rule 3"],
    }


def test_parse_keeps_blank_lines_as_rules():
    # This matches the current implementation, which does *not* skip empty lines.
    text = """\
global 1
global 2
[web]
rule after blank
"""

    config = RulesConfig.parse(text=text)

    assert config.rules == ["global 1", "global 2"]
    assert config.client_rules == {"web": ["rule after blank"]}
